from ..compat import mock, ReasonableBytesIO
from .base import CommandTestCase
from gandi.cli.commands import dns

# disable SSL requests warning for tests
import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()


RESPONSES = {
    'https://dns.api.gandi.net/api/v5/domains': {
        'status': 200,
        'headers': 'application/json',
        'body': [{'domain_href': 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com',  # noqa
                  'domain_records_href': 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/records',  # noqa
                  'fqdn': 'iheartcli.com'},
                 {'domain_href': 'https://dns.api.gandi.net/api/v5/domains/cli.sexy',  # noqa
                  'domain_records_href': 'https://dns.api.gandi.net/api/v5/domains/cli.sexy/records',  # noqa
                  'fqdn': 'cli.sexy'}],
    },
    'https://dns.api.gandi.net/api/v5/domains/iheartcli.com': {
        'status': 200,
        'headers': 'application/json',
        'body': {'domain_href': 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com',  # noqa
                 'domain_keys_href': 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/keys',  # noqa
                 'domain_records_href': 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/records',  # noqa
                 'fqdn': 'iheartcli.com',
                 'zone_href': 'https://dns.api.gandi.net/api/v5/zones/397c514-e7cb-11e6-9429-00163e6dc886',  # noqa
                 'zone_records_href': 'https://dns.api.gandi.net/api/v5/zones/397c514-e7cb-11e6-9429-00163e6dc886/records',  # noqa
                 'zone_uuid': '397c514-e7cb-11e6-9429-00163e6dc886'}
    },
    'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/records?sort_by=rrset_name': {  # noqa
        'status': 200,
        'headers': 'application/json',
        'body': [{'rrset_href': 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/records/%40/A',  # noqa
                  'rrset_name': '@',
                  'rrset_ttl': 10800,
                  'rrset_type': 'A',
                  'rrset_values': ['217.70.184.38']},
                 {'rrset_href': 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/records/%40/MX',  # noqa
                  'rrset_name': '@',
                  'rrset_ttl': 10800,
                  'rrset_type': 'MX',
                  'rrset_values': ['50 fb.mail.gandi.net.', '10 spool.mail.gandi.net.']},  # noqa
                 {'rrset_href': 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/records/blog/CNAME',  # noqa
                  'rrset_name': 'blog',
                  'rrset_ttl': 10800,
                  'rrset_type': 'CNAME',
                  'rrset_values': ['blogs.vip.gandi.net.']},
                 {'rrset_href': 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/records/imap/CNAME',  # noqa
                  'rrset_name': 'imap',
                  'rrset_ttl': 10800,
                  'rrset_type': 'CNAME',
                  'rrset_values': ['access.mail.gandi.net.']},
                 {'rrset_href': 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/records/pop/CNAME',  # noqa
                  'rrset_name': 'pop',
                  'rrset_ttl': 10800,
                  'rrset_type': 'CNAME',
                  'rrset_values': ['access.mail.gandi.net.']},
                 {'rrset_href': 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/records/smtp/CNAME',  # noqa
                  'rrset_name': 'smtp',
                  'rrset_ttl': 10800,
                  'rrset_type': 'CNAME',
                  'rrset_values': ['relay.mail.gandi.net.']},
                 {'rrset_href': 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/records/webmail/CNAME',  # noqa
                  'rrset_name': 'webmail',
                  'rrset_ttl': 10800,
                  'rrset_type': 'CNAME',
                  'rrset_values': ['webmail.gandi.net.']},
                 {'rrset_href': 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/records/www/CNAME',  # noqa
                  'rrset_name': 'www',
                  'rrset_ttl': 10800,
                  'rrset_type': 'CNAME',
                  'rrset_values': ['webredir.vip.gandi.net.']}],
    },
    'https://dns.api.gandi.net/api/v5/dns/rrtypes': {
        'status': 200,
        'headers': 'application/json',
        'body': ['A', 'AAAA', 'CAA', 'CDS', 'CNAME', 'DNAME', 'DS', 'LOC',
                 'MX', 'NS', 'PTR', 'SPF', 'SRV', 'SSHFP', 'TLSA', 'TXT',
                 'WKS'],
    },
    'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/records': {
        'status': 201,
        'headers': 'application/json',
        'body': {'message': 'DNS Record Created'},
    },
    'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/records/blog/CNAME': {  # noqa
        'status': 204,
        'headers': 'application/json',
        'body': {},
    },
    'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/keys': {
        'status': 200,
        'headers': {'content-type': 'application/json',
                    'location': 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/keys/3415833-2314-4a86-ba1c-c3c58608a168'},  # noqa
        'body': [{'algorithm': 13,
                  'algorithm_name': 'ECDSAP256SHA256',
                  'deleted': False,
                  'ds': 'iheartcli.com. 3600 IN DS 5411 13 2 6153c39cfe4ff8673635490515e19f5336f5b7ee9c5ca4572fc44b24a0e794a',  # noqa
                  'flags': 256,
                  'fqdn': 'iheartcli.com',
                  'key_href': 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/keys/3415833-2314-4a86-ba1c-c3c58608a168',  # noqa
                  'status': 'active',
                  'uuid': '3415833-2314-4a86-ba1c-c3c58608a168'},
                 {'algorithm': 13,
                  'algorithm_name': 'ECDSAP256SHA256',
                  'deleted': False,
                  'ds': 'iheartcli.com. 3600 IN DS 43819 13 2 b4e6ed591f28f4a269b9adfaedec836ea0fe63a8f7f5097108297afa5492b70',  # noqa
                  'flags': 256,
                  'fqdn': 'iheartcli.com',
                  'key_href': 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/keys/adaab60-bb17-40ed-a13e-88376fe28c86',  # noqa
                  'status': 'active',
                  'uuid': 'adaab60-bb17-40ed-a13e-88376fe28c86'}],
    },
    'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/keys/3415833-2314-4a86-ba1c-c3c58608a168': {  # noqa
        'status': 200,
        'headers': 'application/json',
        'body': {'algorithm': 13,
                 'algorithm_name': 'ECDSAP256SHA256',
                 'deleted': False,
                 'ds': 'iheartcli.com. 3600 IN DS 5411 13 2 6153c39cfe4ff8673635490515e19f5336f5b7ee9c5ca4572fc44b24a0e794a',  # noqa
                 'flags': 256,
                 'fqdn': 'iheartcli.com',
                 'public_key': 'Gnhra3gcNHUL0d05Ia6F/tgBzDD/Km6c2XFZA9RAOcjk/qg9aodc79MQtsTx4/CBlTmCSRIxlXWm1yMmV3LOlw==',  # noqa
                 'fingerprint': '626168cae12c674f38958b324e10c7bb63ed74cc9d649bf04766a7c095c865787',  # noqa
                 'key_href': 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/keys/3415833-2314-4a86-ba1c-c3c58608a168',  # noqa
                 'status': 'active',
                 'tag': 40658,
                 'uuid': '3415833-2314-4a86-ba1c-c3c58608a168'},
    },
    'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/keys/adaab60-bb17-40ed-a13e-88376fe28c86': {  # noqa
        'status': 204,
        'headers': 'application/json',
        'body': {},
    },
}


def _mock_requests(method, url, *args, **kwargs):
    # print(method, url, args, kwargs)
    content = RESPONSES[url]['body']
    headers = RESPONSES[url]['headers']
    if kwargs.get('headers', {}).get('Accept') == 'text/plain':
        content = """\
@ 10800 IN A 217.70.184.38
@ 10800 IN MX 10 spool.mail.gandi.net.
@ 10800 IN MX 50 fb.mail.gandi.net.
@ 10800 IN SOA ns1.gandi.net. hostmaster.gandi.net. 197539823 10800 3600 604800 10800
blog 10800 IN CNAME blogs.vip.gandi.net.
imap 10800 IN CNAME access.mail.gandi.net.
pop 10800 IN CNAME access.mail.gandi.net.
smtp 10800 IN CNAME relay.mail.gandi.net.
webmail 10800 IN CNAME webmail.gandi.net.
www 10800 IN CNAME webredir.vip.gandi.net."""  # noqa
    content_hdr = kwargs.get('headers', {}).get('Content-Type')
    if method == 'PUT' and content_hdr == 'text/plain':
        content = {'message': 'DNS Record Created'}
    if method == 'PUT' and url == 'https://dns.api.gandi.net/api/v5/domains/iheartcli.com/records/blog/CNAME':  # noqa
        content = {'message': 'DNS Record Created'}
    mock_resp = mock.Mock()
    mock_resp.status_code = 200
    mock_resp.content = content
    mock_resp.headers = headers
    mock_resp.json = mock.Mock(return_value=content)
    return mock_resp


class DnsTestCase(CommandTestCase):

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_domain_list(self, mock_request):
        mock_request.side_effect = _mock_requests

        result = self.invoke_with_exceptions(dns.domain_list, [])

        wanted = """\
iheartcli.com
cli.sexy
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_list(self, mock_request):
        mock_request.side_effect = _mock_requests

        result = self.invoke_with_exceptions(dns.list, ['iheartcli.com'])

        wanted = """\
name        : @
ttl         : 10800
type        : A
values      : 217.70.184.38
----------
name        : @
ttl         : 10800
type        : MX
values      : 50 fb.mail.gandi.net., 10 spool.mail.gandi.net.
----------
name        : blog
ttl         : 10800
type        : CNAME
values      : blogs.vip.gandi.net.
----------
name        : imap
ttl         : 10800
type        : CNAME
values      : access.mail.gandi.net.
----------
name        : pop
ttl         : 10800
type        : CNAME
values      : access.mail.gandi.net.
----------
name        : smtp
ttl         : 10800
type        : CNAME
values      : relay.mail.gandi.net.
----------
name        : webmail
ttl         : 10800
type        : CNAME
values      : webmail.gandi.net.
----------
name        : www
ttl         : 10800
type        : CNAME
values      : webredir.vip.gandi.net.
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_list_filter(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com', '--type', 'CNAME']
        result = self.invoke_with_exceptions(dns.list, args)

        wanted = """\
----------
name        : blog
ttl         : 10800
type        : CNAME
values      : blogs.vip.gandi.net.
----------
name        : imap
ttl         : 10800
type        : CNAME
values      : access.mail.gandi.net.
----------
name        : pop
ttl         : 10800
type        : CNAME
values      : access.mail.gandi.net.
----------
name        : smtp
ttl         : 10800
type        : CNAME
values      : relay.mail.gandi.net.
----------
name        : webmail
ttl         : 10800
type        : CNAME
values      : webmail.gandi.net.
----------
name        : www
ttl         : 10800
type        : CNAME
values      : webredir.vip.gandi.net.
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_list_filter_args(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com', 'blog', 'CNAME']
        result = self.invoke_with_exceptions(dns.list, args)

        wanted = """\
----------
name        : blog
ttl         : 10800
type        : CNAME
values      : blogs.vip.gandi.net.
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_list_unknown(self, mock_request):
        mock_request.side_effect = _mock_requests

        result = self.invoke_with_exceptions(dns.list, ['example.com'])

        wanted = """\
Sorry domain example.com does not exist
Please use one of the following: iheartcli.com, cli.sexy
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_list_text(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com', '--text']
        result = self.invoke_with_exceptions(dns.list, args)

        wanted = """\
@ 10800 IN A 217.70.184.38
@ 10800 IN MX 10 spool.mail.gandi.net.
@ 10800 IN MX 50 fb.mail.gandi.net.
@ 10800 IN SOA ns1.gandi.net. hostmaster.gandi.net. 197539823 10800 3600 604800 10800
blog 10800 IN CNAME blogs.vip.gandi.net.
imap 10800 IN CNAME access.mail.gandi.net.
pop 10800 IN CNAME access.mail.gandi.net.
smtp 10800 IN CNAME relay.mail.gandi.net.
webmail 10800 IN CNAME webmail.gandi.net.
www 10800 IN CNAME webredir.vip.gandi.net.
"""  # noqa
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_create(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com', 'blog', 'cname', 'blog.cli.sexy']
        result = self.invoke_with_exceptions(dns.create, args)

        wanted = """DNS Record Created
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_update_missing(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com']
        result = self.invoke_with_exceptions(dns.update, args)

        wanted = """Cannot find parameters for zone content to update.
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_update_unknown(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['example.com']
        result = self.invoke_with_exceptions(dns.update, args)

        wanted = """\
Sorry domain example.com does not exist
Please use one of the following: iheartcli.com, cli.sexy
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_update_argument_ok(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com', '--file', 'sandbox/example.txt']

        content = """\
blog 10800 IN CNAME blogs.vip.gandi.net.
"""
        result = self.isolated_invoke_with_exceptions(dns.update, args,
                                                      temp_content=content)
        wanted = """DNS Record Created
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_update_parameters_ok(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com', 'blog', 'cname', 'blog.cli.sexy']

        result = self.invoke_with_exceptions(dns.update, args)
        wanted = """DNS Record Created
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_update_parameters_ko(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com', 'blog', 'cname']

        result = self.invoke_with_exceptions(dns.update, args)
        wanted = """You must provide one or more value parameter.
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_update_pipe_ok(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com']

        content = b"""\
blog 10800 IN CNAME blogs.vip.gandi.net.
"""
        result = self.invoke_with_exceptions(dns.update, args,
                                             input=ReasonableBytesIO(content))
        wanted = """DNS Record Created
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_create_multiple(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com', 'blog', 'cname', 'blog.cli.sexy',
                'glop.cli.sexy']
        result = self.invoke_with_exceptions(dns.create, args)

        wanted = """DNS Record Created
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_create_type_case(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com', 'blog', 'CNAME', 'blog.cli.sexy']
        result = self.invoke_with_exceptions(dns.create, args)

        wanted = """DNS Record Created
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_create_unknown(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['example.com', 'blog', 'CNAME', 'blog.cli.sexy']
        result = self.invoke_with_exceptions(dns.create, args)

        wanted = """\
Sorry domain example.com does not exist
Please use one of the following: iheartcli.com, cli.sexy
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_delete(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com', 'blog', 'CNAME', '-f']
        result = self.invoke_with_exceptions(dns.delete, args)

        wanted = """Delete successful.
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_delete_unknown(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['example.com', 'blog', 'CNAME', '-f']
        result = self.invoke_with_exceptions(dns.delete, args)

        wanted = """\
Sorry domain example.com does not exist
Please use one of the following: iheartcli.com, cli.sexy
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_delete_all(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com']
        result = self.invoke_with_exceptions(dns.delete, args, input='\n')

        wanted = """\
Are you sure to delete all records for domain iheartcli.com ? [y/N]: \n"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_delete_name(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com', 'blog']
        result = self.invoke_with_exceptions(dns.delete, args, input='\n')

        wanted = """\
Are you sure to delete all 'blog' name records for domain iheartcli.com ? [y/N]: \n"""  # noqa
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_delete_prompt(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com', 'blog', 'CNAME']
        result = self.invoke_with_exceptions(dns.delete, args, input='\n')

        wanted = """\
Are you sure to delete all 'blog' records of type CNAME for domain iheartcli.com ? [y/N]: \n"""  # noqa
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_keys_list(self, mock_request):
        mock_request.side_effect = _mock_requests

        result = self.invoke_with_exceptions(dns.keys_list, ['iheartcli.com'])
        wanted = """\
uuid           : 3415833-2314-4a86-ba1c-c3c58608a168
algorithm      : 13
algorithm_name : ECDSAP256SHA256
ds             : iheartcli.com. 3600 IN DS 5411 13 2 6153c39cfe4ff8673635490515e19f5336f5b7ee9c5ca4572fc44b24a0e794a
flags          : 256
status         : active
----------
uuid           : adaab60-bb17-40ed-a13e-88376fe28c86
algorithm      : 13
algorithm_name : ECDSAP256SHA256
ds             : iheartcli.com. 3600 IN DS 43819 13 2 b4e6ed591f28f4a269b9adfaedec836ea0fe63a8f7f5097108297afa5492b70
flags          : 256
status         : active
"""  # noqa
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_keys_info(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com', '3415833-2314-4a86-ba1c-c3c58608a168']
        result = self.invoke_with_exceptions(dns.keys_info, args)
        wanted = """\
uuid           : 3415833-2314-4a86-ba1c-c3c58608a168
algorithm      : 13
algorithm_name : ECDSAP256SHA256
ds             : iheartcli.com. 3600 IN DS 5411 13 2 6153c39cfe4ff8673635490515e19f5336f5b7ee9c5ca4572fc44b24a0e794a
fingerprint    : 626168cae12c674f38958b324e10c7bb63ed74cc9d649bf04766a7c095c865787
public_key     : Gnhra3gcNHUL0d05Ia6F/tgBzDD/Km6c2XFZA9RAOcjk/qg9aodc79MQtsTx4/CBlTmCSRIxlXWm1yMmV3LOlw==
flags          : 256
tag            : 40658
status         : active
"""  # noqa
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_keys_create(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com', '256']
        result = self.invoke_with_exceptions(dns.keys_create, args)
        wanted = """\
uuid           : 3415833-2314-4a86-ba1c-c3c58608a168
algorithm      : 13
algorithm_name : ECDSAP256SHA256
ds             : iheartcli.com. 3600 IN DS 5411 13 2 6153c39cfe4ff8673635490515e19f5336f5b7ee9c5ca4572fc44b24a0e794a
fingerprint    : 626168cae12c674f38958b324e10c7bb63ed74cc9d649bf04766a7c095c865787
public_key     : Gnhra3gcNHUL0d05Ia6F/tgBzDD/Km6c2XFZA9RAOcjk/qg9aodc79MQtsTx4/CBlTmCSRIxlXWm1yMmV3LOlw==
flags          : 256
tag            : 40658
status         : active
"""  # noqa
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_keys_delete_ok(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com', 'adaab60-bb17-40ed-a13e-88376fe28c86', '-f']
        result = self.invoke_with_exceptions(dns.keys_delete, args)
        wanted = """Delete successful.
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_keys_delete_prompt(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com', 'adaab60-bb17-40ed-a13e-88376fe28c86']
        result = self.invoke_with_exceptions(dns.keys_delete, args, input='\n')
        wanted = """\
Are you sure you want to delete key adaab60-bb17-40ed-a13e-88376fe28c86 on \
domain iheartcli.com? [y/N]: \n"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)

    @mock.patch('gandi.cli.core.client.requests.request')
    def test_dns_keys_recover(self, mock_request):
        mock_request.side_effect = _mock_requests
        args = ['iheartcli.com', 'adaab60-bb17-40ed-a13e-88376fe28c86']
        result = self.invoke_with_exceptions(dns.keys_recover, args)
        wanted = """Recover successful.
"""
        self.assertEqual(result.output, wanted)
        self.assertEqual(result.exit_code, 0)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import re
import os
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.md')).read()
CHANGES = open(os.path.join(here, 'CHANGES.rst')).read()


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass into py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        try:
            from multiprocessing import cpu_count
            self.pytest_args = ['-n', str(cpu_count()), '--boxed']
            self.pytest_args = []
        except (ImportError, NotImplementedError):
            self.pytest_args = ['-n', '1', '--boxed']

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


with open(os.path.join(here, 'gandi', 'cli', '__init__.py')) as v_file:
    version = re.compile(r".*__version__ = '(.*?)'",
                         re.S).match(v_file.read()).group(1)

requires = ['setuptools', 'pyyaml', 'click>=7.0', 'requests', 'IPy']

tests_require = ['pytest', 'pytest-cov', 'tox']

if sys.version_info < (3, 0):
    tests_require += ['mock']
else:
    if sys.version_info < (3, 4):
        raise RuntimeError(
            "Python 3 earlier than 3.4 is not supported"
            " (sys.version: {})".format(sys.version))

extras_require = {
    'test': tests_require,
}

setup(name='gandi.cli',
      namespace_packages=['gandi'],
      version=version,
      description='Gandi command line interface',
      long_description=README + '\n\n' + CHANGES,
      author='Gandi',
      author_email='feedback@gandi.net',
      classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Terminals',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
      ],
      url='https://github.com/Gandi/gandi.cli',
      packages=find_packages(),
      cmdclass={'test': PyTest},
      include_package_data=True,
      zip_safe=False,
      install_requires=requires,
      tests_require=tests_require,
      extras_require=extras_require,
      entry_points={
          'console_scripts': [
              'gandi = gandi.cli.__main__:main',
          ],
      },
      )



""" API commands module. """

from gandi.cli.core.base import GandiModule


class Api(GandiModule):

    """ Module to handle CLI commands.

    $ gandi api
    """

    @classmethod
    def info(cls):
        """Display information about API."""
        return cls.call('version.info')



""" Disk namespace commands. """

import click
from click.exceptions import UsageError

from gandi.cli.core.cli import cli
from gandi.cli.core.utils import (
    output_disk, output_generic, randomstring,
    DatacenterLimited
)
from gandi.cli.core.utils.size import disk_check_size
from gandi.cli.core.params import (pass_gandi, DATACENTER, SNAPSHOTPROFILE_VM,
                                   KERNEL, SIZE, option, DISK_IMAGE)


@cli.group(name='disk')
@pass_gandi
def disk(gandi):
    """Commands related to hosting disks."""


@disk.command()
@click.option('--only-data', help='Only display data disks.', is_flag=True)
@click.option('--only-snapshot', help='Only display snapshots.', is_flag=True)
@click.option('--attached', help='Only display disks attached to a VM',
              is_flag=True)
@click.option('--detached', help='Only display detached disks', is_flag=True)
@click.option('--type', help='Display types.', is_flag=True)
@click.option('--id', help='Display ids.', is_flag=True)
@click.option('--vm', help='Display vms.', is_flag=True)
@click.option('--snapshotprofile', help='Display snapshot profile.',
              is_flag=True)
@click.option('--datacenter', default=None, type=DATACENTER,
              help='Filter results by datacenter.')
@click.option('--limit', help='Limit number of results.', default=100,
              show_default=True)
@pass_gandi
def list(gandi, only_data, only_snapshot, attached, detached, type, id, vm,
         snapshotprofile, datacenter, limit):
    """ List disks. """
    options = {
        'items_per_page': limit,
    }

    if attached and detached:
        raise UsageError('You cannot use both --attached and --detached.')

    if only_data:
        options.setdefault('type', []).append('data')
    if only_snapshot:
        options.setdefault('type', []).append('snapshot')
    if datacenter:
        options['datacenter_id'] = gandi.datacenter.usable_id(datacenter)

    output_keys = ['name', 'state', 'size']
    if type:
        output_keys.append('type')
    if id:
        output_keys.append('id')
    if vm:
        output_keys.append('vm')

    profiles = []
    if snapshotprofile:
        output_keys.append('profile')
        profiles = gandi.snapshotprofile.list()

    result = gandi.disk.list(options)
    vms = dict([(vm_['id'], vm_) for vm_ in gandi.iaas.list()])

    # filter results per attached/detached
    disks = []
    for disk in result:
        if attached and not disk['vms_id']:
            continue
        if detached and disk['vms_id']:
            continue
        disks.append(disk)

    for num, disk in enumerate(disks):
        if num:
            gandi.separator_line()
        output_disk(gandi, disk, [], vms, profiles, output_keys)

    return result


@disk.command()
@click.argument('resource', nargs=-1, required=True)
@pass_gandi
def info(gandi, resource):
    """ Display information about a disk.

    Resource can be a disk name or ID
    """
    output_keys = ['name', 'state', 'size', 'type', 'id', 'dc', 'vm',
                   'profile', 'kernel', 'cmdline']

    resource = sorted(tuple(set(resource)))
    vms = dict([(vm['id'], vm) for vm in gandi.iaas.list()])
    datacenters = gandi.datacenter.list()

    result = []
    for num, item in enumerate(resource):
        if num:
            gandi.separator_line()
        disk = gandi.disk.info(item)
        output_disk(gandi, disk, datacenters, vms, [], output_keys)
        result.append(disk)

    return result


@disk.command()
@click.option('--background', '--bg', default=False, is_flag=True,
              help='Run command in background mode (default=False).')
@click.option('--force', '-f', is_flag=True,
              help='This is a dangerous option that will cause CLI to continue'
                   ' without prompting. (default=False).')
@pass_gandi
@click.argument('resource', nargs=-1, required=True)
def detach(gandi, resource, background, force):
    """ Detach disks from currectly attached vm.

    Resource can be a disk name, or ID
    """
    resource = sorted(tuple(set(resource)))
    if not force:
        proceed = click.confirm('Are you sure you want to detach %s?' %
                                ', '.join(resource))
        if not proceed:
            return

    result = gandi.disk.detach(resource, background)
    if background:
        gandi.pretty_echo(result)

    return result


@disk.command()
@click.option('--background', '--bg', default=False, is_flag=True,
              help='Run command in background mode (default=False).')
@click.option('-r', '--read-only', default=False, is_flag=True,
              help='Attach disk as read-only')
@click.option('--position', '-p', type=click.INT,
              help='Position where disk should be attached: 0 for system disk.'
                   ' If there is already a disk attached at the specified'
                   ' position, it will be swapped.')
@click.option('--force', '-f', is_flag=True,
              help='This is a dangerous option that will cause CLI to continue'
                   ' without prompting. (default=False).')
@pass_gandi
@click.argument('disk', nargs=1, required=True)
@click.argument('vm', nargs=1, required=True)
def attach(gandi, disk, vm, position, read_only, background, force):
    """ Attach disk to vm.

    disk can be a disk name, or ID
    vm can be a vm name, or ID
    """
    if not force:
        proceed = click.confirm("Are you sure you want to attach disk '%s'"
                                " to vm '%s'?" % (disk, vm))
        if not proceed:
            return

    disk_info = gandi.disk.info(disk)
    attached = disk_info.get('vms_id', False)
    if attached and not force:
        gandi.echo('This disk is still attached')
        proceed = click.confirm('Are you sure you want to detach %s?' % disk)

        if not proceed:
            return

    result = gandi.disk.attach(disk, vm, background, position, read_only)
    if background and result:
        gandi.pretty_echo(result)

    return result


@disk.command()
@click.option('--cmdline', type=click.STRING, default=None,
              help='Kernel cmdline.')
@click.option('--kernel', type=KERNEL, default=None, help='Kernel for disk.')
@click.option('--name', type=click.STRING, default=None, help='Disk name.')
@click.option('--size', default=None, metavar='[+]SIZE[M|G|T]', type=SIZE,
              help=('Disk size. A size suffix (M for megabytes up to T for '
                    'terabytes) is optional, megabytes is the default if no '
                    'suffix is present. A prefix + is optionnal, if provided '
                    'size value will be added to current disk size, default '
                    'is to set directly new disk size.'),
              callback=disk_check_size)
@click.option('--snapshotprofile', help='Selected snapshot profile.',
              default=None, type=SNAPSHOTPROFILE_VM)
@click.option('--delete-snapshotprofile', default=False, is_flag=True,
              help='Remove snapshot profile associated to this disk.')
@click.option('--background', '--bg', default=False, is_flag=True,
              help='Run command in background mode (default=False).')
@pass_gandi
@click.argument('resource')
def update(gandi, resource, cmdline, kernel, name, size,
           snapshotprofile, delete_snapshotprofile, background):
    """ Update a disk.

    Resource can be a disk name, or ID
    """
    if snapshotprofile and delete_snapshotprofile:
        raise UsageError('You must not set snapshotprofile and '
                         'delete-snapshotprofile.')

    if delete_snapshotprofile:
        snapshotprofile = ''

    if kernel:
        source_info = gandi.disk.info(resource)
        available = gandi.kernel.is_available(source_info, kernel)
        if not available:
            raise UsageError('Kernel %s is not available for disk %s' %
                             (kernel, resource))

    result = gandi.disk.update(resource, name, size, snapshotprofile,
                               background, cmdline, kernel)
    if background:
        gandi.pretty_echo(result)

    return result


@disk.command()
@click.option('--force', '-f', is_flag=True,
              help='This is a dangerous option that will cause CLI to continue'
                   ' without prompting. (default=False).')
@click.option('--background', '--bg', default=False, is_flag=True,
              help='run command in background mode (default=False).')
@click.argument('resource', nargs=-1, required=True)
@pass_gandi
def delete(gandi, resource, force, background):
    """ Delete a disk. """
    output_keys = ['id', 'type', 'step']

    resource = sorted(tuple(set(resource)))
    if not force:
        disk_info = "'%s'" % ', '.join(resource)
        proceed = click.confirm('Are you sure you want to delete disk %s?'
                                % disk_info)

        if not proceed:
            return

    opers = gandi.disk.delete(resource, background)
    if background:
        for oper in opers:
            output_generic(gandi, oper, output_keys)

    return opers


@disk.command()
@click.option('--name', type=click.STRING, default=None,
              help='Disk name, will be generated if not provided.')
@click.option('--vm', default=None, type=click.STRING,
              help='Attach the newly created disk to the vm.')
@click.option('--size', default='3072', metavar='SIZE[M|G|T]', type=SIZE,
              help=('Disk size. A size suffix (M for megabytes up to T for '
                    'terabytes) is optional, megabytes is the default if no '
                    'suffix is present.'),
              callback=disk_check_size)
@click.option('--snapshotprofile', help='Selected snapshot profile.',
              default=None, type=SNAPSHOTPROFILE_VM)
@click.option('--source', default=None, type=DISK_IMAGE,
              help='Create a disk from a disk or a snapshot.')
@option('--datacenter', type=DATACENTER, default='FR-SD5',
        help='Datacenter where the disk will be created.')
@click.option('--background', '--bg', default=False, is_flag=True,
              help='Run command in background mode (default=False).')
@pass_gandi
def create(gandi, name, vm, size, snapshotprofile, datacenter, source,
           background):
    """ Create a new disk. """
    try:
        gandi.datacenter.is_opened(datacenter, 'iaas')
    except DatacenterLimited as exc:
        gandi.echo('/!\ Datacenter %s will be closed on %s, '
                   'please consider using another datacenter.' %
                   (datacenter, exc.date))

    if vm:
        vm_dc = gandi.iaas.info(vm)
        vm_dc_id = vm_dc['datacenter_id']
        dc_id = int(gandi.datacenter.usable_id(datacenter))
        if vm_dc_id != dc_id:
            gandi.echo('/!\ VM %s datacenter will be used instead of %s.'
                       % (vm, datacenter))
        datacenter = vm_dc_id

    output_keys = ['id', 'type', 'step']
    name = name or randomstring('vdi')

    disk_type = 'data'
    oper = gandi.disk.create(name, vm, size, snapshotprofile, datacenter,
                             source, disk_type, background)

    if background:
        output_generic(gandi, oper, output_keys)

    return oper


@disk.command()
@click.option('--name', type=click.STRING, default=None,
              help='Snapshot name, will be generated if not provided.')
@click.option('--background', '--bg', default=False, is_flag=True,
              help='Run command in background mode (default=False).')
@click.argument('resource')
@pass_gandi
def snapshot(gandi, name, resource, background):
    """ Create a snapshot on the fly. """
    name = name or randomstring('snp')

    source_info = gandi.disk.info(resource)
    datacenter = source_info['datacenter_id']
    result = gandi.disk.create(name, None, None, None, datacenter, resource,
                               'snapshot', background)

    if background:
        gandi.pretty_echo(result)
    return result


@disk.command()
@click.option('--background', '--bg', default=False, is_flag=True,
              help='Run command in background mode (default=False).')
@click.argument('resource', required=True)
@pass_gandi
def rollback(gandi, resource, background):
    """ Rollback a disk from a snapshot. """
    result = gandi.disk.rollback(resource, background)

    if background:
        gandi.pretty_echo(result)
    return result


@disk.command()
@click.option('--background', '--bg', default=False, is_flag=True,
              help='Run command in background mode (default=False).')
@click.option('--force', '-f', is_flag=True,
              help='This is a dangerous option that will cause CLI to continue'
                   ' without prompting. (default=False).')
@click.argument('resource', required=True)
@pass_gandi
def migrate(gandi, resource, force, background):
    """ Migrate a disk to another datacenter. """
    # check it's not attached
    source_info = gandi.disk.info(resource)
    if source_info['vms_id']:
        click.echo('Cannot start the migration: disk %s is attached. '
                   'Please detach the disk before starting the migration.'
                   % resource)
        return

    disk_datacenter = source_info['datacenter_id']
    dc_choices = gandi.datacenter.list_migration_choice(disk_datacenter)
    if not dc_choices:
        click.echo('No datacenter is available for migration')
        return
    elif len(dc_choices) == 1:
        # use the only one available
        datacenter_id = dc_choices[0]['id']
    else:
        choice_list = [dc['dc_code'] for dc in dc_choices]
        dc_choice = click.Choice(choice_list)
        dc_chosen = click.prompt('Select a datacenter [%s]' % '|'.join(choice_list),  # noqa
                                 type=dc_choice,
                                 show_default=True)
        datacenter_id = [dc['id'] for dc in dc_choices
                         if dc['dc_code'] == dc_chosen][0]

    if not force:
        proceed = click.confirm('Are you sure you want to migrate disk %s ?'
                                % resource)
        if not proceed:
            return

    datacenters = gandi.datacenter.list()
    dc_from = [dc['dc_code']
               for dc in datacenters if dc['id'] == disk_datacenter][0]
    dc_to = [dc['dc_code']
             for dc in datacenters if dc['id'] == datacenter_id][0]
    migration_msg = ('* Starting the migration of disk %s from datacenter %s '
                     'to %s' % (resource, dc_from, dc_to))
    gandi.echo(migration_msg)

    output_keys = ['id', 'type', 'step']
    oper = gandi.disk.migrate(resource, datacenter_id, background)
    if background:
        output_generic(gandi, oper, output_keys)

    return oper



# -*- coding: utf-8 -*-
""" Configuration namespace tests. """

import os

from .base import CommandTestCase
from gandi.cli.commands import config


class ConfigTestCase(CommandTestCase):

    def test_get_empty(self):
        result = self.invoke_with_exceptions(config.get, [])
        self.assertEqual(result.exit_code, 2)

    def test_get_unknown(self):
        result = self.invoke_with_exceptions(config.get, ['unknown-key'])
        self.assertEqual(result.output, """No value found.
""")
        self.assertEqual(result.exit_code, 1)

    def test_get(self):
        result = self.invoke_with_exceptions(config.get, ['api'])
        self.assertEqual(result.exit_code, 0)

    def test_set_empty(self):
        result = self.invoke_with_exceptions(config.set, [])
        self.assertEqual(result.exit_code, 2)

        result = self.invoke_with_exceptions(config.set, ['some-key'])
        self.assertEqual(result.exit_code, 2)

    def test_set_empty_value(self):
        result = self.invoke_with_exceptions(config.set, ['some-key'])
        self.assertEqual(result.exit_code, 2)

    def test_set_get(self):
        result = self.invoke_with_exceptions(config.set, ['dummy'])
        self.assertEqual(result.exit_code, 2)

        result = self.invoke_with_exceptions(config.set, ['dummy',
                                                          'v4lu3'])
        self.assertEqual(result.exit_code, 0)

        result = self.invoke_with_exceptions(config.get, ['dummy'])
        self.assertEqual(result.output, """v4lu3
""")
        self.assertEqual(result.exit_code, 0)

    def test_delete_empty(self):
        result = self.invoke_with_exceptions(config.set, [])
        self.assertEqual(result.exit_code, 2)

    def test_delete(self):
        result = self.invoke_with_exceptions(config.set, ['dummy', 'value'])
        self.assertEqual(result.exit_code, 0)

        result = self.invoke_with_exceptions(config.get, ['dummy'])
        self.assertEqual(result.output, """value
""")
        self.assertEqual(result.exit_code, 0)

        result = self.invoke_with_exceptions(config.delete, ['dummy'])
        self.assertEqual(result.exit_code, 0)

        result = self.invoke_with_exceptions(config.get, ['unknown-key'])
        self.assertEqual(result.output, """No value found.
""")
        self.assertEqual(result.exit_code, 1)

    def test_edit(self):
        os.environ['EDITOR'] = '/usr/bin/nano'
        result = self.invoke_with_exceptions(config.get, ['editor'])
        self.assertEqual(result.output, """/usr/bin/nano
""")
        self.assertEqual(result.exit_code, 0)

        del os.environ['EDITOR']
        result = self.invoke_with_exceptions(config.set, ['editor',
                                                          '/usr/bin/vi'])
        self.assertEqual(result.exit_code, 0)

        result = self.invoke_with_exceptions(config.get, ['editor'])
        self.assertEqual(result.output, """/usr/bin/vi
""")
        self.assertEqual(result.exit_code, 0)



""" Virtual machines namespace commands. """

import click

from gandi.cli.core.cli import cli
from gandi.cli.core.utils import (
    output_vm, output_image, output_generic, output_datacenter,
    output_kernels, output_metric,
    DatacenterLimited
)
from gandi.cli.core.utils.size import disk_check_size
from gandi.cli.core.utils.password import mkpassword
from gandi.cli.core.params import (
    pass_gandi, option, IntChoice, DATACENTER, DISK_IMAGE, SIZE
)


@cli.group(name='vm')
@pass_gandi
def vm(gandi):
    """Commands related to hosting virtual machines."""


@vm.command()
@click.option('--state', default=None, help='Filter results by state.')
@click.option('--datacenter', default=None, type=DATACENTER,
              help='Filter results by datacenter.')
@click.option('--id', help='Display ids.', is_flag=True)
@click.option('--limit', help='Limit number of results.', default=100,
              show_default=True)
@pass_gandi
def list(gandi, state, id, limit, datacenter):
    """List virtual machines."""
    options = {
        'items_per_page': limit,
    }
    if state:
        options['state'] = state
    if datacenter:
        options['datacenter_id'] = gandi.datacenter.usable_id(datacenter)

    output_keys = ['hostname', 'state']
    if id:
        output_keys.append('id')

    result = gandi.iaas.list(options)
    for num, vm in enumerate(result):
        if num:
            gandi.separator_line()
        output_vm(gandi, vm, [], output_keys)

    return result


@vm.command()
@click.argument('resource', nargs=-1, required=True)
@click.option('--stat', default=False, is_flag=True,
              help='Display general vm statistic')
@pass_gandi
def info(gandi, resource, stat):
    """Display information about a virtual machine.

    Resource can be a Hostname or an ID
    """
    output_keys = ['hostname', 'state', 'cores', 'memory', 'console',
                   'datacenter', 'ip']
    justify = 14
    if stat is True:
        sampler = {'unit': 'hours', 'value': 1, 'function': 'max'}
        time_range = 3600 * 24
        query_vif = 'vif.bytes.all'
        query_vbd = 'vbd.bytes.all'

    resource = sorted(tuple(set(resource)))
    datacenters = gandi.datacenter.list()
    ret = []
    for num, item in enumerate(resource):
        if num:
            gandi.separator_line()
        vm = gandi.iaas.info(item)
        output_vm(gandi, vm, datacenters, output_keys, justify)
        ret.append(vm)
        for disk in vm['disks']:
            gandi.echo('')
            disk_out_keys = ['label', 'kernel_version', 'name', 'size']
            output_image(gandi, disk, datacenters, disk_out_keys, justify,
                         warn_deprecated=False)
        if stat is True:
            metrics_vif = gandi.metric.query(vm['id'], time_range, query_vif,
                                             'vm', sampler)
            metrics_vbd = gandi.metric.query(vm['id'], time_range, query_vbd,
                                             'vm', sampler)
            gandi.echo('')
            gandi.echo('vm network stats')
            output_metric(gandi, metrics_vif, 'direction', justify)
            gandi.echo('disk network stats')
            output_metric(gandi, metrics_vbd, 'direction', justify)

    return ret


@vm.command()
@click.option('--background', '--bg', default=False, is_flag=True,
              help='Run command in background mode (default=False).')
@click.argument('resource', nargs=-1, required=True)
@pass_gandi
def stop(gandi, background, resource):
    """Stop a virtual machine.

    Resource can be a Hostname or an ID
    """
    output_keys = ['id', 'type', 'step']

    resource = sorted(tuple(set(resource)))
    opers = gandi.iaas.stop(resource, background)
    if background:
        for oper in opers:
            output_generic(gandi, oper, output_keys)

    return opers


@vm.command()
@click.option('--background', '--bg', default=False, is_flag=True,
              help='Run command in background mode (default=False).')
@click.argument('resource', nargs=-1, required=True)
@pass_gandi
def start(gandi, background, resource):
    """Start a virtual machine.

    Resource can be a Hostname or an ID
    """
    output_keys = ['id', 'type', 'step']

    resource = sorted(tuple(set(resource)))
    opers = gandi.iaas.start(resource, background)
    if background:
        for oper in opers:
            output_generic(gandi, oper, output_keys)

    return opers


@vm.command()
@click.option('--background', '--bg', default=False, is_flag=True,
              help='Run command in background mode (default=False).')
@click.argument('resource', nargs=-1, required=True)
@pass_gandi
def reboot(gandi, background, resource):
    """Reboot a virtual machine.

    Resource can be a Hostname or an ID
    """
    output_keys = ['id', 'type', 'step']

    resource = sorted(tuple(set(resource)))
    opers = gandi.iaas.reboot(resource, background)
    if background:
        for oper in opers:
            output_generic(gandi, oper, output_keys)

    return opers


@vm.command()
@click.option('--background', '--bg', default=False, is_flag=True,
              help='Run command in background mode (default=False).')
@click.option('--force', '-f', is_flag=True,
              help='This is a dangerous option that will cause CLI to continue'
                   ' without prompting. (default=False).')
@click.argument('resource', nargs=-1, required=True)
@pass_gandi
def delete(gandi, background, force, resource):
    """Delete a virtual machine.

    Resource can be a Hostname or an ID
    """
    output_keys = ['id', 'type', 'step']

    resource = sorted(tuple(set(resource)))
    possible_resources = gandi.iaas.resource_list()
    for item in resource:
        if item not in possible_resources:
            gandi.echo('Sorry virtual machine %s does not exist' % item)
            gandi.echo('Please use one of the following: %s' %
                       possible_resources)
            return

    if not force:
        instance_info = "'%s'" % ', '.join(resource)
        proceed = click.confirm("Are you sure to delete Virtual Machine %s?" %
                                instance_info)

        if not proceed:
            return

    iaas_list = gandi.iaas.list()
    stop_opers = []
    for item in resource:
        vm = next((vm for (index, vm) in enumerate(iaas_list)
                   if vm['hostname'] == item), gandi.iaas.info(item))
        if vm['state'] == 'running':
            if background:
                gandi.echo('Virtual machine not stopped, background option '
                           'disabled')
                background = False
            oper = gandi.iaas.stop(item, background)
            if not background:
                stop_opers.append(oper)

    opers = gandi.iaas.delete(resource, background)
    if background:
        for oper in stop_opers + opers:
            output_generic(gandi, oper, output_keys)

    return opers


@vm.command()
@option('--datacenter', type=DATACENTER, default='FR-SD5',
        help='Datacenter where the VM will be spawned.')
@option('--memory', type=click.INT, default=256,
        help='Quantity of RAM in Megabytes to allocate.')
@option('--cores', type=click.INT, default=1,
        help='Number of cpu.')
@click.option('--ip-version', type=IntChoice(['4', '6']), default=None,
              help='Version of created IP.')
@option('--bandwidth', type=click.INT, default=102400,
        help="Network bandwidth in kbit/s used to create the VM's first "
             "network interface.")
@click.option('--login', default=None,
              help='Login to create on the VM.')
@click.option('--password', default=False, is_flag=True,
              help='Will ask for a password to be set for the root account '
                   'and the created login.')
@click.option('--hostname', default=None,
              help='Hostname of the VM, will be generated if not provided.')
@option('--image', type=DISK_IMAGE, default='Debian 8',
        help='Disk image used to boot the VM. A * prefix means the image is '
             'deprecated and will soon be unavailable.')
@click.option('--run', default=None,
              help='Shell command that will run at the first startup of a VM.'
                   'This command will run with root privileges in the ``/`` '
                   'directory at the end of its boot: network interfaces and '
                   'disks are mounted.')
@click.option('--background', '--bg', default=False, is_flag=True,
              help='Run command in background mode (default=False).')
@option('--sshkey', multiple=True,
        help='Authorize ssh authentication for the given ssh key.')
@click.option('--size', default=None, metavar='SIZE[M|G|T]', type=SIZE,
              help=('Disk size. A size suffix (M for megabytes up to T for '
                    'terabytes) is optional, megabytes is the default if no '
                    'suffix is present.'),
              callback=disk_check_size)
@click.option('--vlan', default=None, help='A vlan to use with this vm.')
@click.option('--ip', default=None, help='An ip in the vlan for this vm.')
@click.option('--script', default=None,
              help='Local script to upload and run on the VM after creation.')
@click.option('--script-args', default=None,
              help='Local script argument line.')
@click.option('--ssh', default=False, is_flag=True,
              help='Open a SSH session to the machine after creation '
                   '(default=False).')
@click.option('--gen-password', default=False, is_flag=True,
              help='Generate a random password to be set for the root '
                   'account and the created login(default=False).')
@pass_gandi
def create(gandi, datacenter, memory, cores, ip_version, bandwidth, login,
           password, hostname, image, run, background, sshkey, size, vlan, ip,
           script, script_args, ssh, gen_password):
    """Create a new virtual machine.

    you can specify a configuration entry named 'sshkey' containing
    path to your sshkey file

    $ gandi config set [-g] sshkey ~/.ssh/id_rsa.pub

    or getting the sshkey "my_key" from your gandi ssh keyring

    $ gandi config set [-g] sshkey my_key

    to know which disk image label (or id) to use as image

    $ gandi vm images

    """
    try:
        gandi.datacenter.is_opened(datacenter, 'iaas')
    except DatacenterLimited as exc:
        if exc.date:
            gandi.echo('/!\ Datacenter %s will be closed on %s, '
                       'please consider using another datacenter.' %
                       (datacenter, exc.date))

    if gandi.image.is_deprecated(image, datacenter):
        gandi.echo('/!\ Image %s is deprecated and will soon be unavailable.'
                   % image)

    pwd = None
    if gen_password:
        pwd = mkpassword()

    if password or (not pwd and not sshkey):
        pwd = click.prompt('password', hide_input=True,
                           confirmation_prompt=True)

    if ip and not vlan:
        gandi.echo("--ip can't be used without --vlan.")
        return

    if not vlan and not ip_version:
        ip_version = 6

    if not ip_version:
        gandi.echo("* Private only ip vm (can't enable emergency web console "
                   'access).')

    # Display a short summary for creation
    if login:
        user_summary = 'root and %s users' % login
        password_summary = 'Users root and %s' % login
    else:
        user_summary = 'root user'
        password_summary = 'User root'

    gandi.echo('* %s will be created.' % user_summary)
    if sshkey:
        gandi.echo('* SSH key authorization will be used.')
    if pwd and gen_password:
        gandi.echo('* %s setup with password %s' %
                   (password_summary, pwd))
    if not pwd:
        gandi.echo('* No password supplied for vm (required to enable '
                   'emergency web console access).')
    result = gandi.iaas.create(datacenter, memory, cores, ip_version,
                               bandwidth, login, pwd, hostname,
                               image, run,
                               background,
                               sshkey, size, vlan, ip, script, script_args,
                               ssh)
    if background:
        gandi.echo('* IAAS backend is now creating your VM and its '
                   'associated resources in the background.')

    return result


@vm.command()
@click.option('--memory', type=click.INT, default=None,
              help='Quantity of RAM in Megabytes to allocate.')
@click.option('--cores', type=click.INT, default=None,
              help='Number of cpu.')
@click.option('--console', default=None, is_flag=True,
              help='Activate the emergency console.')
@click.option('--password', default=False, is_flag=True,
              help='Will ask for a password to be set for the root account '
                   'and the created login.')
@click.option('--background', '--bg', default=False, is_flag=True,
              help='Run command in background mode (default=False).')
@click.option('--reboot', default=False, is_flag=True,
              help='Accept a VM reboot for non-live updates')
@click.argument('resource')
@pass_gandi
def update(gandi, resource, memory, cores, console, password, background,
           reboot):
    """Update a virtual machine.

    Resource can be a Hostname or an ID
    """
    pwd = None
    if password:
        pwd = click.prompt('password', hide_input=True,
                           confirmation_prompt=True)

    max_memory = None
    if memory:
        max_memory = gandi.iaas.required_max_memory(resource, memory)

    if max_memory and not reboot:
        gandi.echo('memory update must be done offline.')
        if not click.confirm("reboot machine %s?" % resource):
            return

    result = gandi.iaas.update(resource, memory, cores, console, pwd,
                               background, max_memory)
    if background:
        gandi.pretty_echo(result)

    return result


@vm.command()
@click.argument('resource')
@pass_gandi
def console(gandi, resource):
    """Open a console to virtual machine.

    Resource can be a Hostname or an ID
    """
    gandi.echo('/!\ Please be aware that if you didn\'t provide a password '
               'during creation, console service will be unavailable.')
    gandi.echo('/!\ You can use "gandi vm update" command to set a password.')
    gandi.echo('/!\ Use ~. ssh escape key to exit.')

    gandi.iaas.console(resource)


@vm.command()
@click.option('--wait', default=False, is_flag=True,
              help='Wait for virtual machine sshd to come up (timeout 2min).')
@click.option('--wipe-key', default=False, is_flag=True,
              help='Wipe SSH known host entry first.')
@click.option('--login', '-l', default='root',
              help='Use given login for ssh call')
@click.option('--identity', '-i', default=None,
              help='Use specified path for ssh key')
@click.argument('resource')
@click.argument('args', nargs=-1)
@pass_gandi
def ssh(gandi, resource, login, identity, wipe_key, wait, args):
    """Spawn an SSH session to virtual machine.

    Resource can be a Hostname or an ID
    """
    if '@' in resource:
        (login, resource) = resource.split('@', 1)
    if wipe_key:
        gandi.iaas.ssh_keyscan(resource)
    if wait:
        gandi.iaas.wait_for_sshd(resource)
    gandi.iaas.ssh(resource, login, identity, args)


@vm.command()
@click.option('--datacenter', type=DATACENTER, default=None,
              help='Filter by datacenter.')
@click.argument('label', required=False)
@pass_gandi
def images(gandi, label, datacenter):
    """List available system images for virtual machines.

    You can also filter results using label, by example:

    $ gandi vm images Ubuntu --datacenter LU

    or

    $ gandi vm images 'Ubuntu 10.04' --datacenter LU

    """
    output_keys = ['label', 'os_arch', 'kernel_version', 'disk_id',
                   'dc', 'name']

    datacenters = gandi.datacenter.list()
    result = gandi.image.list(datacenter, label)
    for num, image in enumerate(result):
        if num:
            gandi.separator_line()
        output_image(gandi, image, datacenters, output_keys)

    # also display usable disks
    result = gandi.disk.list_create(datacenter, label)
    for disk in result:
        gandi.separator_line()
        output_image(gandi, disk, datacenters, output_keys)

    return result


@vm.command()
@click.option('--vm', default=None,
              help='Output available kernels for given vm.')
@click.option('--datacenter', type=DATACENTER, default=None,
              help='Filter by datacenter.')
@click.option('--flavor', default=None,
              help='Filter by kernel flavor.')
@click.argument('match', default='', required=False, metavar='pattern')
@pass_gandi
def kernels(gandi, vm, datacenter, flavor, match):
    """List available kernels."""

    if vm:
        vm = gandi.iaas.info(vm)

    dc_list = gandi.datacenter.filtered_list(datacenter, vm)

    for num, dc in enumerate(dc_list):
        if num:
            gandi.echo('\n')
        output_datacenter(gandi, dc, ['dc_name'])
        kmap = gandi.kernel.list(dc['id'], flavor, match)
        for _flavor in kmap:
            gandi.separator_line()
            output_kernels(gandi, _flavor, kmap[_flavor])


@cli.command()
@click.option('--id', help='Display ids.', is_flag=True)
@pass_gandi
def datacenters(gandi, id):
    """List available datacenters."""
    output_keys = ['iso', 'name', 'country', 'dc_code', 'status']
    if id:
        output_keys.append('id')

    result = gandi.datacenter.list()
    for num, dc in enumerate(result):
        if num:
            gandi.separator_line()
        output_datacenter(gandi, dc, output_keys, justify=10)

    return result


@vm.command()
@click.option('--background', '--bg', default=False, is_flag=True,
              help='Run command in background mode (default=False).')
@click.option('--force', '-f', is_flag=True,
              help='This is a dangerous option that will cause CLI to continue'
                   ' without prompting. (default=False).')
@click.option('--finalize', default=False, is_flag=True,
              help='Finalize migration.')
@click.argument('resource', required=True)
@pass_gandi
def migrate(gandi, resource, force, background, finalize):
    """ Migrate a virtual machine to another datacenter. """
    if not gandi.iaas.check_can_migrate(resource):
        return

    if not force:
        proceed = click.confirm('Are you sure you want to migrate VM %s ?'
                                % resource)
        if not proceed:
            return

    if finalize:
        gandi.iaas.need_finalize(resource)

    output_keys = ['id', 'type', 'step']
    oper = gandi.iaas.migrate(resource, background, finalize=finalize)
    if background:
        output_generic(gandi, oper, output_keys)

    return oper



""" Record namespace commands. """

import os
import click
import json

from gandi.cli.core.cli import cli
from gandi.cli.core.utils import (
    output_generic,
)
from gandi.cli.core.params import pass_gandi, StringConstraint


@cli.group(name='record')
@pass_gandi
def record(gandi):
    """Commands related to DNS zone records (xmlrpc)."""


@record.command()
@click.option('--zone-id', '-z', default=None, type=click.INT,
              help='Zone ID to use, if not set default zone will be used.')
@click.option('--output', '-o', is_flag=True,
              help='Write the records into a file.')
@click.option('--format', '-f', type=click.Choice(['text', 'json']),
              help='Choose the output format', required=False)
@click.option('--limit', help='Limit number of results.', default=100,
              show_default=True)
@click.argument('domain', required=True)
@pass_gandi
def list(gandi, domain, zone_id, output, format, limit):
    """List DNS zone records for a domain."""
    options = {
        'items_per_page': limit,
    }
    output_keys = ['name', 'type', 'value', 'ttl']

    if not zone_id:
        result = gandi.domain.info(domain)
        zone_id = result['zone_id']

    if not zone_id:
        gandi.echo('No zone records found, domain %s doesn\'t seems to be '
                   'managed at Gandi.' % domain)
        return

    records = gandi.record.list(zone_id, options)

    if not output and not format:
        for num, rec in enumerate(records):
            if num:
                gandi.separator_line()
            output_generic(gandi, rec, output_keys, justify=12)
    elif output:
        zone_filename = domain + "_" + str(zone_id)
        if os.path.isfile(zone_filename):
            open(zone_filename, 'w').close()
        for record in records:
            format_record = ('%s %s IN %s %s' %
                             (record['name'], record['ttl'],
                              record['type'], record['value']))
            with open(zone_filename, 'ab') as zone_file:
                zone_file.write(format_record + '\n')
        gandi.echo('Your zone file have been writen in %s' % zone_filename)
    elif format:
        if format == 'text':
            for record in records:
                format_record = ('%s %s IN %s %s' %
                                 (record['name'], record['ttl'],
                                  record['type'], record['value']))
                gandi.echo(format_record)
        if format == 'json':
            format_record = json.dumps(records, sort_keys=True,
                                       indent=4, separators=(',', ': '))
            gandi.echo(format_record)

    return records


@record.command()
@click.option('--zone-id', '-z', default=None, type=click.INT,
              help='Zone ID to use, if not set, default zone will be used.')
@click.option('--name', default=None, required=True,
              help='Relative name, may contain leading wildcard. '
                   '`@` for empty name')
@click.option('--type', default=None, required=True,
              type=click.Choice(['A', 'AAAA', 'CNAME', 'MX', 'NS', 'TXT',
                                 'WKS', 'SRV', 'LOC', 'SPF']),
              help='DNS record type')
@click.option('--value', default=None, required=True,
              type=StringConstraint(minlen=1, maxlen=1024),
              help='Value for record. Semantics depends on the record type.'
                   'Currently limited to 1024 ASCII characters.'
                   'In case of TXT, each part between quotes is limited to 255'
                   ' characters')
@click.option('--ttl', default=None, required=False,
              type=click.IntRange(min=300, max=2592000),
              help='Time to live, in seconds, between 5 minutes and 30 days')
@click.argument('domain', required=True)
@pass_gandi
def create(gandi, domain, zone_id, name, type, value, ttl):
    """Create new DNS zone record entry for a domain."""

    if not zone_id:
        result = gandi.domain.info(domain)
        zone_id = result['zone_id']

    if not zone_id:
        gandi.echo('No zone records found, domain %s doesn\'t seems to be '
                   'managed at Gandi.' % domain)
        return

    record = {'type': type, 'name': name, 'value': value}
    if ttl:
        record['ttl'] = ttl

    result = gandi.record.create(zone_id, record)
    return result


@record.command()
@click.option('--zone-id', '-z', default=None, type=click.INT,
              help='Zone ID to use, if not set, default zone will be used.')
@click.option('--name', default=None,
              help='Relative name of the record to delete.')
@click.option('--type', default=None,
              type=click.Choice(['A', 'AAAA', 'CNAME', 'MX', 'NS', 'TXT',
                                 'WKS', 'SRV', 'LOC', 'SPF']),
              help='DNS record type')
@click.option('--value', default=None,
              type=StringConstraint(minlen=1, maxlen=1024),
              help='Value for record. Semantics depends on the record type.'
                   'Currently limited to 1024 ASCII characters.'
                   'In case of TXT, each part between quotes is limited to 255'
                   ' characters')
@click.argument('domain', required=True)
@pass_gandi
def delete(gandi, domain, zone_id, name, type, value):
    """Delete a record entry for a domain"""
    if not zone_id:
        result = gandi.domain.info(domain)
        zone_id = result['zone_id']

    if not zone_id:
        gandi.echo('No zone records found, domain %s doesn\'t seems to be '
                   'managed at Gandi.' % domain)
        return
    if not name and not type and not value:
        proceed = click.confirm('This command without parameters --type, '
                                '--name or --value will remove all records'
                                ' in this zone file. Are you sur to '
                                'perform this action ?')
        if not proceed:
            return
    record = {'name': name, 'type': type, 'value': value}
    result = gandi.record.delete(zone_id, record)
    return result


@record.command()
@click.option('--zone-id', '-z', default=None, type=click.INT,
              help='Zone ID to use, if not set, default zone will be used.')
@click.option('--file', '-f', type=click.File('r'),
              required=False, help='Filename of the zone file.')
@click.option('--record', '-r', default=None, required=False,
              help="'name TTL IN TYPE [A, AAAA, MX, TXT, SPF] value' \n"
              "Note that you can specify only a name, but in case of "
              "multiple entries with the same name, only the first one will "
              "be updated")
@click.option('--new-record', default=None, required=False,
              help="'name TTL IN TYPE [A, AAAA, MX, TXT, SPF] value'")
@click.argument('domain', required=True)
@pass_gandi
def update(gandi, domain, zone_id, file, record, new_record):
    """Update records entries for a domain.

    You can update an individual record using --record and --new-record
    parameters

    Or you can use a plaintext file to update all records of a DNS zone at
    once with --file parameter.
    """
    if not zone_id:
        result = gandi.domain.info(domain)
        zone_id = result['zone_id']

    if not zone_id:
        gandi.echo('No zone records found, domain %s doesn\'t seems to be'
                   ' managed at Gandi.' % domain)
        return
    if file:
        records = file.read()
        result = gandi.record.zone_update(zone_id, records)
        return result
    elif record and new_record:
        result = gandi.record.update(zone_id, record, new_record)
        return result
    else:
        gandi.echo('You must indicate a zone file or a record.'
                   ' Use `gandi record update --help` for more information')



"""Size related methods."""

import click
from gandi.cli.core.cli import compatcallback


@compatcallback
def disk_check_size(ctx, param, value):
    """ Validation callback for disk size parameter."""
    if value:
        # if we've got a prefix
        if isinstance(value, tuple):
            val = value[1]
        else:
            val = value
        if val % 1024:
            raise click.ClickException('Size must be a multiple of 1024.')
    return value



""" Account namespace commands. """

from gandi.cli.core.cli import cli
from gandi.cli.core.utils import output_account
from gandi.cli.core.params import pass_gandi


@cli.group(name='account')
@pass_gandi
def account(gandi):
    """Commands related to accounts."""


@account.command()
@pass_gandi
def info(gandi):
    """Display information about hosting account.
    """
    output_keys = ['handle', 'credit', 'prepaid']

    account = gandi.account.all()
    account['prepaid_info'] = gandi.contact.balance().get('prepaid', {})
    output_account(gandi, account, output_keys)
    return account



""" Domain namespace commands. """

import click

from gandi.cli.core.cli import cli
from gandi.cli.core.params import pass_gandi, DNSSEC_FLAGS, DNSSEC_ALGORITHM


@cli.group(name='dnssec')
@pass_gandi
def dnssec(gandi):
    """Commands related to dnssec."""


@dnssec.command()
@click.option('--flags', default=None, type=DNSSEC_FLAGS,
              help='Flags (ZSK or KSK)')
@click.option('--algorithm', default=None, type=DNSSEC_ALGORITHM,
              help='Algorithm')
@click.option('--public_key', default=None, help='Public key (base64-encoded)')
@click.argument('resource')
@pass_gandi
def create(gandi, resource, flags, algorithm, public_key):
    """Create DNSSEC key."""

    result = gandi.dnssec.create(resource, flags, algorithm, public_key)

    return result


@dnssec.command()
@click.argument('resource')
@pass_gandi
def list(gandi, resource):
    """List DNSSEC keys."""
    keys = gandi.dnssec.list(resource)
    gandi.pretty_echo(keys)

    return keys


@dnssec.command()
@click.argument('resource', type=click.INT)
@pass_gandi
def delete(gandi, resource):
    """Delete DNSSEC key.
    """

    result = gandi.dnssec.delete(resource)
    gandi.echo('Delete successful.')

    return result



""" Gandi CLI core modules.

Contains:
- XML-RPC connection handler
- Configuration handler
- Base class, which load modules and commands
- Custom command line validation parameters
- Output methods
"""



""" PaaS commands module. """

import re
import sys

from gandi.cli.core.base import GandiModule
from gandi.cli.modules.metric import Metric
from gandi.cli.modules.vhost import Vhost
from gandi.cli.modules.datacenter import Datacenter
from gandi.cli.modules.sshkey import SshkeyHelper


class Paas(GandiModule, SshkeyHelper):

    """ Module to handle CLI commands.

    $ gandi paas attach
    $ gandi paas clone
    $ gandi paas console
    $ gandi paas create
    $ gandi paas delete
    $ gandi paas info
    $ gandi paas list
    $ gandi paas restart
    $ gandi paas types
    $ gandi paas update

    """

    @classmethod
    def type_list(cls, options=None):
        """List type of PaaS instances."""
        return cls.safe_call('paas.type.list', options)

    @classmethod
    def clone(cls, name, vhost, directory, origin):
        """Clone a PaaS instance's vhost into a local git repository."""
        paas_info = cls.info(name)

        if 'php' in paas_info['type'] and not vhost:
            cls.error('PHP instances require indicating the VHOST to clone '
                      'with --vhost <vhost>')

        paas_access = '%s@%s' % (paas_info['user'], paas_info['git_server'])
        remote_url = 'ssh+git://%s/%s.git' % (paas_access, vhost)

        command = 'git clone %s %s --origin %s' \
                  % (remote_url, directory, origin)

        init_git = cls.execute(command)
        if init_git:
            cls.echo('Use `git push %s master` to push your code to the '
                     'instance.' % (origin))
            cls.echo('Then `$ gandi deploy` to build and deploy your '
                     'application.')

    @classmethod
    def attach(cls, name, vhost, remote_name):
        """Attach an instance's vhost to a remote from the local repository."""
        paas_access = cls.get('paas_access')

        if not paas_access:
            paas_info = cls.info(name)
            paas_access = '%s@%s' \
                          % (paas_info['user'], paas_info['git_server'])

        remote_url = 'ssh+git://%s/%s.git' % (paas_access, vhost)

        ret = cls.execute('git remote add %s %s' % (remote_name, remote_url,))

        if ret:
            cls.echo('Added remote `%s` to your local git repository.'
                     % (remote_name))
            cls.echo('Use `git push %s master` to push your code to the '
                     'instance.' % (remote_name))
            cls.echo('Then `$ gandi deploy` to build and deploy your '
                     'application.')

    @classmethod
    def deploy(cls, remote_name, branch):
        """Deploy a PaaS instance."""
        def get_remote_url(remote):
            return 'git config --local --get remote.%s.url' % (remote)

        remote_url = cls.exec_output(get_remote_url(remote_name)) \
            .replace('\n', '')

        if not remote_url or not re.search('gpaas.net|gandi.net', remote_url):
            remote_name = ('$(git config --local --get branch.%s.remote)' %
                           branch)
            remote_url = cls.exec_output(get_remote_url(remote_name)) \
                .replace('\n', '')

        error = None

        if not remote_url:
            error = True
            cls.echo('Error: Could not find git remote '
                     'to extract deploy url from.')
        elif not re.search('gpaas.net|gandi.net', remote_url):
            error = True
            cls.echo('Error: %s is not a valid Simple Hosting git remote.'
                     % (remote_url))
        if error:
            cls.echo("""This usually happens when:
- the current directory has no Simple Hosting git remote attached,
  in this case, please see $ gandi paas attach --help
- the local branch being deployed hasn't been pushed to the \
remote repository yet,
  in this case, please try $ git push <remote> %s
""" % (branch))
            cls.echo('Otherwise, it\'s recommended to use'
                     ' the --remote and/or --branch options:\n'
                     '$ gandi deploy --remote <remote> [--branch <branch>]')
            sys.exit(2)

        remote_url_no_protocol = remote_url.split('://')[1]
        splitted_url = remote_url_no_protocol.split('/')

        paas_access = splitted_url[0]
        deploy_git_host = splitted_url[1]

        command = "ssh %s 'deploy %s %s'" \
                  % (paas_access, deploy_git_host, branch)

        cls.execute(command)

    @classmethod
    def list(cls, options=None):
        """List PaaS instances."""
        return cls.call('paas.list', options)

    @classmethod
    def info(cls, id):
        """Display information about a PaaS instance."""
        return cls.call('paas.info', cls.usable_id(id))

    @classmethod
    def quota(cls, id):
        """return disk quota used/free"""
        sampler = {'unit': 'minutes', 'value': 1, 'function': 'avg'}
        query = 'vfs.df.bytes.all'
        metrics = Metric.query(id, 60, query, 'paas', sampler)

        df = {'free': 0, 'used': 0}
        for metric in metrics:
            what = metric['size'].pop()
            # we need the most recent points
            metric['points'].reverse()
            for point in metric['points']:
                if 'value' in point:
                    df[what] = point['value']
                    break
        return df

    @classmethod
    def cache(cls, id):
        """return the number of query cache for the last 24H"""
        sampler = {'unit': 'days', 'value': 1, 'function': 'sum'}
        query = 'webacc.requests.cache.all'
        metrics = Metric.query(id, 60 * 60 * 24, query, 'paas', sampler)

        cache = {'hit': 0, 'miss': 0, 'not': 0, 'pass': 0}
        for metric in metrics:
            what = metric['cache'].pop()
            for point in metric['points']:
                value = point.get('value', 0)
                cache[what] += value
        return cache

    @classmethod
    def delete(cls, resources, background=False):
        """Delete a PaaS instance."""
        if not isinstance(resources, (list, tuple)):
            resources = [resources]

        opers = []
        for item in resources:
            oper = cls.call('paas.delete', cls.usable_id(item))
            if isinstance(oper, list):
                opers.extend(oper)
            else:
                opers.append(oper)

        if background:
            return opers

        # interactive mode, run a progress bar
        cls.echo('Deleting your PaaS instance.')
        cls.display_progress(opers)

    @classmethod
    def update(cls, id, name, size, quantity, password, sshkey, upgrade,
               console, snapshot_profile, reset_mysql_password, background):
        """Update a PaaS instance."""
        if not background and not cls.intty():
            background = True

        paas_params = {}

        if name:
            paas_params['name'] = name

        if size:
            paas_params['size'] = size

        if quantity:
            paas_params['quantity'] = quantity

        if password:
            paas_params['password'] = password

        paas_params.update(cls.convert_sshkey(sshkey))

        if upgrade:
            paas_params['upgrade'] = upgrade

        if console:
            paas_params['console'] = console

        # XXX to delete a snapshot_profile the value has to be an empty string
        if snapshot_profile is not None:
            paas_params['snapshot_profile'] = snapshot_profile

        if reset_mysql_password:
            paas_params['reset_mysql_password'] = reset_mysql_password

        result = cls.call('paas.update', cls.usable_id(id), paas_params)
        if background:
            return result

        # interactive mode, run a progress bar
        cls.echo('Updating your PaaS instance.')
        cls.display_progress(result)

    @classmethod
    def create(cls, name, size, type, quantity, duration, datacenter, vhosts,
               password, snapshot_profile, background, sshkey):
        """Create a new PaaS instance."""
        if not background and not cls.intty():
            background = True

        datacenter_id_ = int(Datacenter.usable_id(datacenter))

        paas_params = {
            'name': name,
            'size': size,
            'type': type,
            'duration': duration,
            'datacenter_id': datacenter_id_,
        }

        if password:
            paas_params['password'] = password

        if quantity:
            paas_params['quantity'] = quantity

        paas_params.update(cls.convert_sshkey(sshkey))

        if snapshot_profile:
            paas_params['snapshot_profile'] = snapshot_profile

        result = cls.call('paas.create', paas_params)

        if not background:
            # interactive mode, run a progress bar
            cls.echo('Creating your PaaS instance.')
            cls.display_progress(result)
            cls.echo('Your PaaS instance %s has been created.' % name)

        if vhosts:
            paas_info = cls.info(name)
            Vhost.create(paas_info, vhosts, True, background)

        return result

    @classmethod
    def restart(cls, resources, background=False):
        """Restart a PaaS instance."""
        if not isinstance(resources, (list, tuple)):
            resources = [resources]

        opers = []
        for item in resources:
            oper = cls.call('paas.restart', cls.usable_id(item))
            if isinstance(oper, list):
                opers.extend(oper)
            else:
                opers.append(oper)

        if background:
            return opers

        # interactive mode, run a progress bar
        cls.echo('Restarting your PaaS instance.')
        cls.display_progress(opers)

    @classmethod
    def resource_list(cls):
        """ Get the possible list of resources (name, id and vhosts). """
        items = cls.list({'items_per_page': 500})
        ret = [paas['name'] for paas in items]
        ret.extend([str(paas['id']) for paas in items])
        for paas in items:
            paas = cls.info(paas['id'])
            ret.extend([vhost['name'] for vhost in paas['vhosts']])

        return ret

    @classmethod
    def console(cls, id):
        """Open a console to a PaaS instance."""
        oper = cls.call('paas.update', cls.usable_id(id), {'console': 1})
        cls.echo('Activation of the console on your PaaS instance')
        cls.display_progress(oper)
        console_url = Paas.info(cls.usable_id(id))['console']
        access = 'ssh %s' % console_url
        cls.execute(access)

    @classmethod
    def usable_id(cls, id):
        """ Retrieve id from input which can be hostname, vhost, id."""
        try:
            # id is maybe a hostname
            qry_id = cls.from_hostname(id)
            if not qry_id:
                # id is maybe a vhost
                qry_id = cls.from_vhost(id)
            if not qry_id:
                qry_id = int(id)
        except Exception:
            qry_id = None

        if not qry_id:
            msg = 'unknown identifier %s' % id
            cls.error(msg)

        return qry_id

    @classmethod
    def from_vhost(cls, vhost):
        """Retrieve paas instance id associated to a vhost."""
        result = Vhost().list()
        paas_hosts = {}
        for host in result:
            paas_hosts[host['name']] = host['paas_id']

        return paas_hosts.get(vhost)

    @classmethod
    def from_hostname(cls, hostname):
        """Retrieve paas instance id associated to a host."""
        result = cls.list({'items_per_page': 500})
        paas_hosts = {}
        for host in result:
            paas_hosts[host['name']] = host['id']

        return paas_hosts.get(hostname)

    @classmethod
    def list_names(cls):
        """Retrieve paas id and names."""
        ret = dict([(item['id'], item['name'])
                    for item in cls.list({'items_per_page': 500})])
        return ret



""" Webaccelerator commands module """

from gandi.cli.core.base import GandiModule
from gandi.cli.modules.datacenter import Datacenter


class Webacc(GandiModule):

    """ Module to handle CLI commands.

    $ gandi webacc list
    $ gandi webacc info
    $ gandi webacc create
    $ gandi webacc add
    $ gandi webacc delete
    $ gandi webacc enable
    $ gandi webacc disable
    $ gandi webacc probe

    """

    @classmethod
    def list(cls, options=None):
        """ List all webaccelerator """
        if not options:
            options = {}
        return cls.call('hosting.rproxy.list', options)

    @classmethod
    def info(cls, id):
        """ Get information about a Webaccelerator """
        return cls.call('hosting.rproxy.info', cls.usable_id(id))

    @classmethod
    def create(cls, name, datacenter, backends, vhosts, algorithm,
               ssl_enable, zone_alter):
        """ Create a webaccelerator """
        datacenter_id_ = int(Datacenter.usable_id(datacenter))
        params = {
            'datacenter_id': datacenter_id_,
            'name': name,
            'lb': {'algorithm': algorithm},
            'override': True,
            'ssl_enable': ssl_enable,
            'zone_alter': zone_alter
        }
        if vhosts:
            params['vhosts'] = vhosts
        if backends:
            params['servers'] = backends
        try:
            result = cls.call('hosting.rproxy.create', params)
            cls.echo('Creating your webaccelerator %s' % params['name'])
            cls.display_progress(result)
            cls.echo('Your webaccelerator have been created')
            return result
        except Exception as err:
            if err.code == 580142:
                for vhost in params['vhosts']:
                    dns_entry = cls.call(
                        'hosting.rproxy.vhost.get_dns_entries',
                        {'datacenter': datacenter_id_, 'vhost': vhost})
                    txt_record = "@ 3600 IN TXT \"%s=%s\"" % (dns_entry['key'],
                                                              dns_entry['txt'])

                    cname_record = "%s 3600 IN CNAME %s" % (dns_entry['key'],
                                                            dns_entry['cname'])

                    cls.echo('The domain %s don\'t use Gandi DNS or you have'
                             ' not sufficient right to alter the zone file. '
                             'Edit your zone file adding this TXT and CNAME '
                             'record and try again :' % vhost)
                    cls.echo(txt_record)
                    cls.echo(cname_record)
                    cls.echo('\nOr add a file containing %s at :\n'
                             'http://%s/%s.txt\n' % (dns_entry['txt'],
                                                     dns_entry['domain'],
                                                     dns_entry['txt']))
                    cls.separator_line('-', 4)
            else:
                cls.echo(err)

    @classmethod
    def update(cls, resource, new_name, algorithm, ssl_enable, ssl_disable):
        """ Update a webaccelerator"""
        params = {}
        if new_name:
            params['name'] = new_name
        if algorithm:
            params['lb'] = {'algorithm': algorithm}
        if ssl_enable:
            params['ssl_enable'] = ssl_enable
        if ssl_disable:
            params['ssl_enable'] = False

        result = cls.call('hosting.rproxy.update', cls.usable_id(resource),
                          params)
        cls.echo('Updating your webaccelerator')
        cls.display_progress(result)
        cls.echo('The webaccelerator have been udated')
        return result

    @classmethod
    def delete(cls, name):
        """ Delete a webaccelerator """
        result = cls.call('hosting.rproxy.delete', cls.usable_id(name))
        cls.echo('Deleting your webaccelerator named %s' % name)
        cls.display_progress(result)
        cls.echo('Webaccelerator have been deleted')
        return result

    @classmethod
    def backend_list(cls, options):
        """ List all servers used by webaccelerator """
        return cls.call('hosting.rproxy.server.list', options)

    @classmethod
    def backend_add(cls, name, backend):
        """ Add a backend into a webaccelerator """
        oper = cls.call(
            'hosting.rproxy.server.create', cls.usable_id(name), backend)
        cls.echo('Adding backend %s:%s into webaccelerator' %
                 (backend['ip'], backend['port']))
        cls.display_progress(oper)
        cls.echo('Backend added')
        return oper

    @classmethod
    def backend_remove(cls, backend):
        """ Remove a backend on a webaccelerator """
        server = cls.backend_list(backend)
        if server:
            oper = cls.call('hosting.rproxy.server.delete', server[0]['id'])
            cls.echo('Removing backend %s:%s into webaccelerator' %
                     (backend['ip'], backend['port']))
            cls.display_progress(oper)
            cls.echo('Your backend have been removed')
            return oper
        else:
            return cls.echo('No backend found')

    @classmethod
    def backend_enable(cls, backend):
        """ Enable a backend for a server """
        server = cls.backend_list(backend)
        if server:
            oper = cls.call('hosting.rproxy.server.enable', server[0]['id'])
            cls.echo('Activating backend %s' % server[0]['ip'])
            cls.display_progress(oper)
            cls.echo('Backend activated')
            return oper
        else:
            return cls.echo('No backend found')

    @classmethod
    def backend_disable(cls, backend):
        """ Disable a backend for a server """
        server = cls.backend_list(backend)
        oper = cls.call('hosting.rproxy.server.disable',
                        server[0]['id'])
        cls.echo('Desactivating backend on server %s' %
                 server[0]['ip'])
        cls.display_progress(oper)
        cls.echo('Backend desactivated')
        return oper

    @classmethod
    def vhost_list(cls):
        """ List all vhosts used by webaccelerator """
        return cls.call('hosting.rproxy.vhost.list')

    @classmethod
    def vhost_add(cls, resource, params):
        """ Add a vhost into a webaccelerator """
        try:
            oper = cls.call(
                'hosting.rproxy.vhost.create', cls.usable_id(resource), params)
            cls.echo('Adding your virtual host (%s) into %s' %
                     (params['vhost'], resource))
            cls.display_progress(oper)
            cls.echo('Your virtual host habe been added')
            return oper
        except Exception as err:
            if err.code == 580142:
                dc = cls.info(resource)
                dns_entry = cls.call('hosting.rproxy.vhost.get_dns_entries',
                                     {'datacenter': dc['datacenter']['id'],
                                      'vhost': params['vhost']})
                txt_record = "%s 3600 IN TXT \"%s=%s\"" % (dns_entry['key'],
                                                           dns_entry['key'],
                                                           dns_entry['txt'])

                cname_record = "%s 3600 IN CNAME %s" % (dns_entry['key'],
                                                        dns_entry['cname'])

                cls.echo('The domain don\'t use Gandi DNS or you have not'
                         ' sufficient right to alter the zone file. '
                         'Edit your zone file adding this TXT and CNAME '
                         'record and try again :')
                cls.echo(txt_record)
                cls.echo(cname_record)
                cls.echo('\nOr add a file containing %s at :\n'
                         'http://%s/%s.txt\n' % (dns_entry['txt'],
                                                 dns_entry['domain'],
                                                 dns_entry['txt']))

            else:
                cls.echo(err)

    @classmethod
    def vhost_remove(cls, name):
        """ Delete a vhost in a webaccelerator """
        oper = cls.call('hosting.rproxy.vhost.delete', name)
        cls.echo('Deleting your virtual host %s' % name)
        cls.display_progress(oper)
        cls.echo('Your virtual host have been removed')
        return oper

    @classmethod
    def probe(cls, resource, enable, disable, test, host, interval,
              http_method, http_response, threshold, timeout, url, window):
        """ Set a probe for a webaccelerator """
        params = {
            'host': host,
            'interval': interval,
            'method': http_method,
            'response': http_response,
            'threshold': threshold,
            'timeout': timeout,
            'url': url,
            'window': window
        }
        if enable:
            params['enable'] = True
        elif disable:
            params['enable'] = False
        if test:
            result = cls.call(
                'hosting.rproxy.probe.test', cls.usable_id(resource), params)
        else:
            result = cls.call(
                'hosting.rproxy.probe.update', cls.usable_id(resource), params)
            cls.display_progress(result)
        return result

    @classmethod
    def probe_enable(cls, resource):
        """ Activate a probe on a webaccelerator """
        oper = cls.call('hosting.rproxy.probe.enable', cls.usable_id(resource))
        cls.echo('Activating probe on %s' % resource)
        cls.display_progress(oper)
        cls.echo('The probe have been activated')
        return oper

    @classmethod
    def probe_disable(cls, resource):
        """ Disable a probe on a webaccelerator """
        oper = cls.call('hosting.rproxy.probe.disable',
                        cls.usable_id(resource))
        cls.echo('Desactivating probe on %s' % resource)
        cls.display_progress(oper)
        cls.echo('The probe have been desactivated')
        return oper

    @classmethod
    def usable_id(cls, id):
        """ Retrieve id from input which can be hostname, vhost, id. """
        try:
            # id is maybe a hostname
            qry_id = cls.from_name(id)
            if not qry_id:
                # id is maybe an ip
                qry_id = cls.from_ip(id)
            if not qry_id:
                qry_id = cls.from_vhost(id)
        except Exception:
            qry_id = None

        if not qry_id:
            msg = 'unknown identifier %s' % id
            cls.error(msg)

        return qry_id

    @classmethod
    def from_name(cls, name):
        """Retrieve webacc id associated to a webacc name."""
        result = cls.list({'items_per_page': 500})
        webaccs = {}
        for webacc in result:
            webaccs[webacc['name']] = webacc['id']
        return webaccs.get(name)

    @classmethod
    def from_ip(cls, ip):
        """Retrieve webacc id associated to a webacc ip"""
        result = cls.list({'items_per_page': 500})
        webaccs = {}
        for webacc in result:
            for server in webacc['servers']:
                webaccs[server['ip']] = webacc['id']
        return webaccs.get(ip)

    @classmethod
    def from_vhost(cls, vhost):
        """Retrieve webbacc id associated to a webacc vhost"""
        result = cls.list({'items_per_page': 500})
        webaccs = {}
        for webacc in result:
            for vhost in webacc['vhosts']:
                webaccs[vhost['name']] = webacc['id']
        return webaccs.get(vhost)



""" Forward namespace commands. """

import click

from gandi.cli.core.cli import cli
from gandi.cli.core.utils import output_forward
from gandi.cli.core.params import pass_gandi, EMAIL_TYPE


@cli.group(name='forward')
@pass_gandi
def forward(gandi):
    """Commands related to domain mail forwards."""


@forward.command()
@click.option('--limit', help='Limit number of results.',
              default=100, show_default=True)
@click.argument('domain', metavar='domain.tld')
@pass_gandi
def list(gandi, domain, limit):
    """List mail forwards for a domain."""
    options = {'items_per_page': limit}
    result = gandi.forward.list(domain, options)
    for forward in result:
        output_forward(gandi, domain, forward)
    return result


@forward.command()
@click.option('--destination', '-d', help='Add forward destination.',
              multiple=True, required=True)
@click.argument('address', type=EMAIL_TYPE, metavar='address@domain.tld')
@pass_gandi
def create(gandi, address, destination):
    """Create a domain mail forward."""
    source, domain = address

    result = gandi.forward.create(domain, source, destination)

    return result


@forward.command()
@click.option('--dest-add', '-a', help='Add forward destination.',
              multiple=True, required=False)
@click.option('--dest-del', '-d', help='Delete forward destination.',
              multiple=True, required=False)
@click.argument('address', type=EMAIL_TYPE, metavar='address@domain.tld')
@pass_gandi
def update(gandi, address, dest_add, dest_del):
    """Update a domain mail forward."""
    source, domain = address

    if not dest_add and not dest_del:
        gandi.echo('Nothing to update: you must provide destinations to '
                   'update, use --dest-add/-a or -dest-del/-d parameters.')
        return

    result = gandi.forward.update(domain, source, dest_add, dest_del)

    return result


@forward.command()
@click.option('--force', '-f', is_flag=True,
              help='This is a dangerous option that will cause CLI to continue'
                   ' without prompting. (default=False).')
@click.argument('address', type=EMAIL_TYPE, metavar='address@domain.tld')
@pass_gandi
def delete(gandi, address, force):
    """Delete a domain mail forward."""
    source, domain = address

    if not force:
        proceed = click.confirm('Are you sure to delete the domain '
                                'mail forward %s@%s ?' % (source, domain))

        if not proceed:
            return

    result = gandi.forward.delete(domain, source)

    return result



""" SSH key commands module. """

import os
from gandi.cli.core.base import GandiModule
from gandi.cli.core.utils import DuplicateResults


class Sshkey(GandiModule):

    """ Module to handle CLI commands.

    $ gandi sshkey create
    $ gandi sshkey delete
    $ gandi sshkey info
    $ gandi sshkey list

    """

    @classmethod
    def from_name(cls, name):
        """Retrieve a sshkey id associated to a name."""
        sshkeys = cls.list({'name': name})
        if len(sshkeys) == 1:
            return sshkeys[0]['id']
        elif not sshkeys:
            return

        raise DuplicateResults('sshkey name %s is ambiguous.' % name)

    @classmethod
    def usable_id(cls, id):
        """ Retrieve id from input which can be name or id."""
        try:
            # id is maybe a sshkey name
            qry_id = cls.from_name(id)
            if not qry_id:
                qry_id = int(id)
        except DuplicateResults as exc:
            cls.error(exc.errors)
        except Exception:
            qry_id = None

        if not qry_id:
            msg = 'unknown identifier %s' % id
            cls.error(msg)

        return qry_id

    @classmethod
    def list(cls, options=None):
        """ List ssh keys."""
        options = options if options else {}
        return cls.call('hosting.ssh.list', options)

    @classmethod
    def info(cls, id):
        """ Display information about an ssh key. """
        return cls.call('hosting.ssh.info', cls.usable_id(id))

    @classmethod
    def create(cls, name, value):
        """ Create a new ssh key."""
        sshkey_params = {
            'name': name,
            'value': value,
        }

        result = cls.call('hosting.ssh.create', sshkey_params)
        return result

    @classmethod
    def delete(cls, id):
        """Delete this ssh key."""
        return cls.call('hosting.ssh.delete', cls.usable_id(id))


class SshkeyHelper(object):

    """ Helper class to handle sshkey configuration entry. """

    @classmethod
    def convert_sshkey(cls, sshkey):
        """ Return dict param with valid entries for vm/paas methods. """
        params = {}
        if sshkey:
            params['keys'] = []
            for ssh in sshkey:
                if os.path.exists(os.path.expanduser(ssh)):
                    if 'ssh_key' in params:
                        cls.echo("Can't have more than one sshkey file.")
                        continue
                    with open(ssh) as fdesc:
                        sshkey_ = fdesc.read()
                    if sshkey_:
                        params['ssh_key'] = sshkey_
                else:
                    sshkey_id = Sshkey.usable_id(ssh)
                    if sshkey_id:
                        params['keys'].append(sshkey_id)
                    else:
                        cls.echo('This is not a ssh key %s' % ssh)

            if not params['keys']:
                params.pop('keys')

        return params



try:
    # python3
    from xmlrpc.client import DateTime
except ImportError:
    # python2
    from xmlrpclib import DateTime

type_list = list


def list(options):

    ret = [{'date_created': DateTime('20150915T18:29:16'),
            'date_start': None,
            'date_updated': DateTime('20150915T18:29:17'),
            'errortype': None,
            'eta': -1863666,
            'id': 100100,
            'cert_id': None,
            'infos': {'extras': {},
                      'id': '',
                      'label': 'iheartcli.com',
                      'product_action': 'renew',
                      'product_name': 'com',
                      'product_type': 'domain',
                      'quantity': ''},
            'last_error': None,
            'params': {'auth_id': 99999999,
                       'current_year': 2015,
                       'domain': 'iheartcli.com',
                       'domain_id': 1234567,
                       'duration': 1,
                       'param_type': 'domain',
                       'remote_addr': '127.0.0.1',
                       'session_id': 2920674,
                       'tld': 'com',
                       'tracker_id': '621cb9f4-472d-4cc1-b4b9-b18cc61e2914'},
            'session_id': 2920674,
            'source': 'PXP561-GANDI',
            'step': 'BILL',
            'type': 'domain_renew'},
           {'date_created': DateTime('20150505T00:00:00'),
            'date_start': None,
            'date_updated': DateTime('20150505T00:00:00'),
            'errortype': None,
            'eta': 0,
            'id': 100200,
            'cert_id': None,
            'infos': {'extras': {},
                      'id': '',
                      'label': '',
                      'product_action': 'billing_prepaid_add_money',
                      'product_name': '',
                      'product_type': 'corporate',
                      'quantity': ''},
            'last_error': None,
            'params': {'amount': 50.0,
                       'auth_id': 99999999,
                       'param_type': 'prepaid_add_money',
                       'prepaid_id': 100000,
                       'remote_addr': '127.0.0.1',
                       'tracker_id': 'ab0e5e67-6ca7-4afc-8311-f20080f15cf1'},
            'session_id': 9844958,
            'source': 'PXP561-GANDI',
            'step': 'BILL',
            'type': 'billing_prepaid_add_money'},
           {'step': 'RUN',
            'cert_id': 710,
            'id': 100300,
            'type': 'certificate_update',
            'params': {'cert_id': 710,
                       'param_type': 'certificate_update',
                       'prepaid_id': 100000,
                       'inner_step': 'comodo_oper_updated',
                       'dcv_method': 'email',
                       'csr': '-----BEGIN CERTIFICATE REQUEST-----'
                              'MIICxjCCAa4CAQAwgYAxCzAJBgNVBAYTAkZSMQsw'
                              '0eWfyJJTOypoToCtdGoye507GOsgIysfRWaExay5'
                              '-----END CERTIFICATE REQUEST-----',
                       'remote_addr': '127.0.0.1'}},
           {'step': 'RUN',
            'cert_id': 706,
            'id': 100302,
            'type': 'certificate_update',
            'params': {'cert_id': 706,
                       'param_type': 'certificate_update',
                       'prepaid_id': 100000,
                       'inner_step': 'comodo_oper_updated',
                       'dcv_method': 'dns',
                       'csr': '-----BEGIN CERTIFICATE REQUEST-----'
                              'MIICxjCCAa4CAQAwgYAxCzAJBgNVBAYTAkZSMQsw'
                              '0eWfyJJTOypoToCtdGoye507GOsgIysfRWaExay5'
                              '-----END CERTIFICATE REQUEST-----',
                       'remote_addr': '127.0.0.1'}},
           {'step': 'WAIT',
            'cert_id': 701,
            'id': 100303,
            'type': 'certificate_update',
            'params': {'cert_id': 706,
                       'param_type': 'certificate_update',
                       'prepaid_id': 100000,
                       'inner_step': 'check_email_sent',
                       'dcv_method': 'dns',
                       'remote_addr': '127.0.0.1'}},
           {'step': 'RUN',
            'id': 99001,
            'vm_id': 152967,
            'type': 'hosting_migration_vm',
            'params': {'inner_step': 'wait_sync'}},
           {'step': 'RUN',
            'id': 99002,
            'vm_id': 152966,
            'type': 'hosting_migration_vm',
            'params': {'inner_step': 'wait_finalize'}},
           ]

    options.pop('sort_by', None)
    options.pop('items_per_page', None)

    def compare(op, option):
        if isinstance(option, (type_list, tuple)):
            return op in option
        return op == option

    for fkey in options:
        ret = [op for op in ret if compare(op.get(fkey), options[fkey])]

    return ret


def info(id):

    if id == 200:
        return {'step': 'DONE'}

    if id == 300:
        return {'step': 'DONE', 'vm_id': 9000}

    if id == 400:
        return {'step': 'DONE'}

    if id == 600:
        return {'step': 'DONE', 'type': 'certificate_update',
                'params': {'cert_id': 710,
                           'param_type': 'certificate_update',
                           'prepaid_id': 100000,
                           'remote_addr': '127.0.0.1'}}

    if id == 9900:
        return {'step': 'DONE',
                'type': 'hosting_migration_disk',
                'params': {'to_dc_id': 3,
                           'from_dc_id': 1,
                           'inner_step': 'wait_finalize'},
                'id': 9900}

    return [oper for oper in list({}) if oper['id'] == id][0]



import logging
import importlib

log = logging.getLogger(__name__)


class Api(object):

    _calls = {}

    def request(self, method, apikey, *args, **kwargs):
        log.info('Calling %s%r %r' % (method, args, kwargs))
        modname, func = method.split('.', 1)
        modname = 'gandi.cli.tests.fixtures._' + modname
        module = importlib.import_module(modname)
        func = func.replace('.', '_')
        if (kwargs.get('dry_run', False)
                and kwargs.get('return_dry_run', False)):
            func = func + '_dry_run'
        try:
            self._calls.setdefault(method, []).append(args)
            return getattr(module, func)(*args)
        except Exception as exc:
            log.exception('Unexpected Exception %s while calling %s' % (exc,
                                                                        method)
                          )



""" Snapshot profiles namespace commands. """

import click

from gandi.cli.core.cli import cli
from gandi.cli.core.utils import output_snapshot_profile
from gandi.cli.core.params import pass_gandi


@cli.group(name='snapshotprofile')
@pass_gandi
def snapshotprofile(gandi):
    """Commands related to hosting snapshot profiles."""


@snapshotprofile.command()
@click.option('--only-paas', help='Only display PaaS profiles.', is_flag=True)
@click.option('--only-vm', help='Only display vm profile.s', is_flag=True)
@pass_gandi
def list(gandi, only_paas, only_vm):
    """ List snapshot profiles. """
    target = None
    if only_paas and not only_vm:
        target = 'paas'
    if only_vm and not only_paas:
        target = 'vm'

    output_keys = ['id', 'name', 'kept_total', 'target']
    result = gandi.snapshotprofile.list({}, target=target)

    for num, profile in enumerate(result):
        if num:
            gandi.separator_line()
        output_snapshot_profile(gandi, profile, output_keys)

    return result


@snapshotprofile.command()
@click.argument('resource')
@pass_gandi
def info(gandi, resource):
    """ Display information about a snapshot profile.

    Resource can be a profile name or ID
    """
    output_keys = ['id', 'name', 'kept_total', 'target', 'quota_factor',
                   'schedules']

    result = gandi.snapshotprofile.info(resource)
    output_snapshot_profile(gandi, result, output_keys)
    return result
